package cn.com.qmth.stmms.ms.marking.service;

import cn.com.qmth.stmms.ms.commons.config.SystemConfig;
import cn.com.qmth.stmms.ms.core.domain.CheckData;
import cn.com.qmth.stmms.ms.core.domain.Level;
import cn.com.qmth.stmms.ms.core.domain.MarkSubject;
import cn.com.qmth.stmms.ms.core.domain.Paper;
import cn.com.qmth.stmms.ms.core.domain.enums.CheckDataStatusEnum;
import cn.com.qmth.stmms.ms.core.domain.enums.CheckItemEnum;
import cn.com.qmth.stmms.ms.core.domain.user.CheckDataError;
import cn.com.qmth.stmms.ms.core.repository.*;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import javax.persistence.criteria.Predicate;
import java.io.File;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

@Service
public class CheckDataService {

    private static final String SEPARATE = "##";

    @Resource
    SystemConfig systemConfig;

    @Resource
    CheckDataRepo checkDataRepo;

    @Resource
    MarkSubjectRepo markSubjectRepo;

    @Resource
    PaperRepo paperRepo;

    @Resource
    LevelRepo levelRepo;

    @Resource
    CheckDataErrorRepo checkDataErrorRepo;

    /**
     * 异步任务做检查
     *
     * @param checkData 任务对象
     */
    @Async
    public void syncCheckData(CheckData checkData) {
        // 执行中
        checkData.setStatus(CheckDataStatusEnum.RUNNING);
        checkDataRepo.saveAndFlush(checkData);

        int errorCount = 0;
        List<CheckDataError> list = null;
        String content = null;
        try {
            Long checkDataId = checkData.getId();
            Long workId = checkData.getWorkId();
            if (CheckItemEnum.PAPER_SIZE.equals(checkData.getCheckItem())) {
                errorCount = checkPaperSize(workId);
                if (errorCount > 0) {
                    content = String.format("有%d张图片大小为0", errorCount);
                }
            } else if (CheckItemEnum.ALL_SCORE.equals(checkData.getCheckItem())) {
                Map<String, Long> map = checkAllScore(workId);
                if (map != null) {
                    errorCount = map.values().stream().mapToInt(Math::toIntExact).sum();
                }
                if (errorCount > 0) {
                    StringJoiner stringJoiner = new StringJoiner("；");
                    for (Map.Entry<String, Long> entry : map.entrySet()) {
                        stringJoiner.add(String.format("%s有%d张没有成绩", entry.getKey(), entry.getValue()));
                    }
                    content = stringJoiner.toString();
                }
            } else if (CheckItemEnum.LEVEL_SCORE_MATCH.equals(checkData.getCheckItem())) {
                list = checkLevelScoreMatch(checkDataId, workId);
                errorCount = list.size();
            } else if (CheckItemEnum.LEVEL_DIFF.equals(checkData.getCheckItem())) {
                list = checkDiff(checkDataId, workId, checkData.getParamValue(), true);
                errorCount = list.size();
            } else if (CheckItemEnum.SCORE_DIFF.equals(checkData.getCheckItem())) {
                list = checkDiff(checkDataId, workId, checkData.getParamValue(), false);
                errorCount = list.size();
            } else {
                throw new RuntimeException("没有此类型任务");
            }
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            checkData.setStatus(CheckDataStatusEnum.FINISH);
            checkData.setResult(errorCount == 0);
            checkData.setErrorCount(errorCount);
            checkData.setContent(content);
            checkData.setEndTime(new Date());
            saveData(checkData, list);
        }
    }

    @Transactional
    public void saveData(CheckData checkData, List<CheckDataError> list) {
        checkDataRepo.saveAndFlush(checkData);

        // 清除
        checkDataErrorRepo.deleteByCheckDataId(checkData.getId());
        // 插入新数据
        if (list != null && !list.isEmpty()) {
            List<CheckDataError> data = new ArrayList<>();
            //2000条提交一次
            for (CheckDataError checkDataError : list) {
                if (data.size() == 2000) {
                    checkDataErrorRepo.save(data);
                    data.clear();
                }
                data.add(checkDataError);
            }
            //将剩下的数据也导入
            if (!data.isEmpty()) {
                checkDataErrorRepo.save(data);
            }
        }
    }

    /**
     * 检查图片大小0KB
     *
     * @param workId 工作ID
     */
    private int checkPaperSize(Long workId) {
        int count = 0;
        // image
        File imageDir = new File(systemConfig.getImageDir() + File.separator + workId);
        count = getZeroCount(count, imageDir);

        // thumb
        File thumbDir = new File(systemConfig.getThumbDir() + File.separator + workId);
        count = getZeroCount(count, thumbDir);

        // sheet
        File sheetDir = new File(systemConfig.getSheetDir() + File.separator + workId);
        count = getZeroCount(count, sheetDir);
        return count;
    }

    private int getZeroCount(int count, File fileDir) {
        List<File> files = getAllFile(fileDir);
        if (!files.isEmpty()) {
            int zeroCount = Math.toIntExact(files.stream().filter(m -> m.length() == 0).count());
            count = count + zeroCount;
        }
        return count;
    }

    /**
     * 校验每张试卷都有分数
     *
     * @param workId 工作ID
     */
    private Map<String, Long> checkAllScore(Long workId) {
        List<Paper> paperList = paperRepo.findByWorkIdAndIsMissing(workId, false);
        if (paperList != null && !paperList.isEmpty()) {
            return paperList.stream().filter(m -> m.getScore() == null).collect(Collectors.groupingBy(m -> m.getSubject().name(), Collectors.counting()));
        }
        return null;
    }

    /**
     * 每张试卷分数都在其所属档位分数区间内
     *
     * @param workId 工作ID
     */
    private List<CheckDataError> checkLevelScoreMatch(Long checkDataId, Long workId) {
        List<MarkSubject> markSubjects = markSubjectRepo.findAllByWorkIdAndEnableTrue(workId);
        List<Paper> paperList = paperRepo.findByWorkIdAndIsMissing(workId, false);
        List<CheckDataError> list = new ArrayList<>();
        if (paperList != null && !paperList.isEmpty()) {
            // 多科目，先按准考证号+考生姓名分组
            Map<String, List<Paper>> collect = paperList.stream().collect(Collectors.groupingBy(m -> m.getExamNumber() + SEPARATE + m.getStudentName()));
            List<Level> levels = levelRepo.findByWorkId(workId);
            Map<String, Level> levelMap = levels.stream().collect(Collectors.toMap(Level::getCode, Function.identity()));
            for (Map.Entry<String, List<Paper>> entry : collect.entrySet()) {
                String key = entry.getKey();
                String[] strings = key.split(SEPARATE);
                List<Paper> papers = entry.getValue();
                StringJoiner stringJoiner = new StringJoiner("；");
                // 每个科目做比较
                List<String> subjectsAll = markSubjects.stream().map(m -> m.getSubject().name()).collect(Collectors.toList());
                papers.forEach(paper -> {
                    String subject = paper.getSubject().name();
                    subjectsAll.remove(subject);
                    // 未分档
                    if (paper.getLevel() == null) {
                        stringJoiner.add(subject + "未分档");
                        return;
                    }
                    // 未打分
                    if (paper.getScore() == null) {
                        stringJoiner.add(subject + "未打分");
                        return;
                    }
                    Level level = levelMap.get(paper.getLevel());
                    if (level == null) {
                        stringJoiner.add("档位" + paper.getLevel() + "不存在");
                        return;
                    }
                    // 已分档并打分，进行校验
                    if (paper.getScore().intValue() <= level.getMinScore() || paper.getScore().intValue() >= level.getMaxScore()) {
                        stringJoiner.add(String.format("%s档位[%s]分数[%s]不匹配", subject, paper.getLevel(), paper.getScore()));
                    }
                });

                // 未采集
                if (!subjectsAll.isEmpty()) {
                    for (String subject : subjectsAll) {
                        stringJoiner.add(subject + "未采集");
                    }
                }

                if (stringJoiner.toString().length() > 1) {
                    list.add(new CheckDataError(checkDataId, strings[0], strings[1], stringJoiner.toString()));
                }
            }
        }
        return list;
    }

    /**
     * 同一考生各科目档位差
     *
     * @param workId     工作ID
     * @param paramValue 变量参数值
     * @param isLevel    是否校验档位
     */
    private List<CheckDataError> checkDiff(Long checkDataId, Long workId, int paramValue, boolean isLevel) {
        List<MarkSubject> markSubjects = markSubjectRepo.findAllByWorkIdAndEnableTrue(workId);
        Map<String, Level> levelMap = new HashMap<>();
        if (isLevel) {
            List<Level> levels = levelRepo.findByWorkId(workId);
            levelMap = levels.stream().collect(Collectors.toMap(Level::getCode, Function.identity()));
        }

        List<Paper> paperList = paperRepo.findByWorkIdAndIsMissing(workId, false);
        Map<String, List<Paper>> collect = paperList.stream().collect(Collectors.groupingBy(m -> m.getExamNumber() + SEPARATE + m.getStudentName()));
        List<CheckDataError> list = new ArrayList<>();
        for (Map.Entry<String, List<Paper>> entry : collect.entrySet()) {
            String key = entry.getKey();
            String[] strings = key.split(SEPARATE);
            List<Paper> papers = entry.getValue();
            // 只考部分科目，校验不通过
            if (papers.size() != markSubjects.size()) {
                list.add(new CheckDataError(checkDataId, strings[0], strings[1], "只考" + papers.size() + "科，无法校验"));
            } else {
                long noLevel = papers.stream().filter(m -> isLevel ? m.getLevel() == null : m.getScore() == null).count();
                if (noLevel > 0) {
                    list.add(new CheckDataError(checkDataId, strings[0], strings[1], "有" + noLevel + "科未" + (isLevel ? "分档" : "打分") + "，无法校验"));
                } else {
                    Map<String, Level> finalLevelMap = levelMap;
                    Map<String, Integer> integersMap = papers.stream().collect(Collectors.toMap(m -> m.getSubject().name(), m -> isLevel ? finalLevelMap.get(m.getLevel()).getLevelValue() : m.getScore().intValue()));
                    // 校验详情
                    Map<String[], Integer> map = calcDiff(integersMap, paramValue);
                    if (!map.isEmpty()) {
                        StringJoiner stringJoiner = new StringJoiner("；");
                        for (Map.Entry<String[], Integer> integerEntry : map.entrySet()) {
                            stringJoiner.add(String.format("%s相差%d", Arrays.toString(integerEntry.getKey()), integerEntry.getValue()));
                        }
                        list.add(new CheckDataError(checkDataId, strings[0], strings[1], stringJoiner.toString()));
                    }
                }
            }
        }
        return list;
    }

    /**
     * 计算差值
     *
     * @param integersMap 值
     * @param paramValue  变量
     * @return true为通过，false为不通过
     */
    private Map<String[], Integer> calcDiff(Map<String, Integer> integersMap, int paramValue) {
        Map<String[], Integer> map = new HashMap<>();
        List<String> stringList = new ArrayList<>(integersMap.keySet());
        for (int i = 0; i < stringList.size(); i++) {
            for (int j = i; j < stringList.size(); j++) {
                if (i != j) {
                    int diffValue = Math.abs(integersMap.get(stringList.get(i)) - integersMap.get(stringList.get(j)));
                    if (diffValue > paramValue) {
                        String[] key = {stringList.get(i), stringList.get(j)};
                        map.put(key, diffValue);
                    }
                }
            }
        }
        return map;
    }

    /**
     * 获取指定文件夹下所有文件，不含文件夹
     *
     * @param dirFile 文件夹
     */
    public static List<File> getAllFile(File dirFile) {
        List<File> files = new ArrayList<>();
        // 如果文件夹不存在或着不是文件夹，则返回 null
        if (Objects.isNull(dirFile) || !dirFile.exists() || dirFile.isFile())
            return files;

        File[] childrenFiles = dirFile.listFiles();
        if (Objects.isNull(childrenFiles) || childrenFiles.length == 0)
            return files;

        for (File childFile : childrenFiles) {

            // 如果时文件，直接添加到结果集合
            if (childFile.isFile()) {
                files.add(childFile);
            } else {
                // 如果是文件夹，则将其内部文件添加进结果集合
                List<File> cFiles = getAllFile(childFile);
                if (cFiles.isEmpty()) {
                    continue;
                }
                files.addAll(cFiles);
            }
        }
        return files;
    }

    public Page<CheckDataError> pageError(Long checkDataId, Pageable pageable) {
        Specification<CheckDataError> specification = (root, query, builder) -> {
            List<Predicate> predicates = new ArrayList<>();
            predicates.add(builder.equal(root.get("checkDataId"), checkDataId));
            return builder.and(predicates.toArray(new Predicate[predicates.size()]));
        };
        Sort sort = new Sort("examNumber");
        Pageable pageable1 = new PageRequest(pageable.getPageNumber(), pageable.getPageSize(), sort);
        return checkDataErrorRepo.findAll(specification, pageable1);

    }
}
