package cn.com.qmth.stmms.ms.marking.service;

import cn.com.qmth.stmms.ms.core.cache.ParamCache;
import cn.com.qmth.stmms.ms.core.domain.Level;
import cn.com.qmth.stmms.ms.core.domain.RoughLevel;
import cn.com.qmth.stmms.ms.core.domain.task.MarkTaskLevel;
import cn.com.qmth.stmms.ms.core.domain.task.MarkTaskRoughLevel;
import cn.com.qmth.stmms.ms.core.domain.user.MarkUser;
import cn.com.qmth.stmms.ms.core.repository.LevelRepo;
import cn.com.qmth.stmms.ms.core.repository.MarkUserRepo;
import cn.com.qmth.stmms.ms.marking.service.arbitration.ArbitrationService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.DoubleStream;
import java.util.stream.Stream;

/**
 * 定档
 * Created by zhengmin on 2016/9/22.
 */
@Service
public class DetermineRoughLevelService {

    public enum DeterType {
        /**
         * 多数
         */
        MAJORITY,
        /**
         * 权重
         */
        WEIGHT,
        /**
         * 仲裁
         */
        ARBITRATE,
        /**
         * 打回
         */
        REJECT
    }

    @Autowired
    private MarkUserRepo markUserRepo;

    /**
     * 计算定档值（规则：1.不是过半定档，取权重，2.是过半定档，没有取优，取过半定档，3.是过半定档，是取优，取过半定档和权重最优值）
     *
     * @param workId    工作ID
     * @param markTasks 任务集合
     */
    public RoughDeterResult determine(Long workId, List<Level> levels, List<MarkTaskRoughLevel> markTasks) {
        List<RoughLevel> roughLevels = parseRoughLevels(levels);
        // 是否过半定档（1：是）
        boolean majority = Optional.ofNullable(ParamCache.roughLevelConfigMap.get(String.valueOf(String.valueOf(workId))).getMajority()).orElse(1) == 1;
        RoughDeterResult majorityResult = null;
        RoughDeterResult weightResult;
        if (majority) {
            MarkTaskRoughLevel[] tasks = markTasks.toArray(new MarkTaskRoughLevel[0]);
            majorityResult = calcMajority(tasks);
        }

        if (majorityResult != null) {
            // 是否取优（过半定档和权重，取最优值）
            boolean takeBest = Optional.ofNullable(ParamCache.roughLevelConfigMap.get(String.valueOf(String.valueOf(workId))).getTakeBest()).orElse(1) == 1;
            if (takeBest) {
                Map<String, String> levelMap = roughLevels.stream().collect(Collectors.toMap(RoughLevel::getRoughCode, RoughLevel::getRoughCode));

                weightResult = calcWeight(workId, roughLevels, markTasks);
                // 取优
                int majorityLevelValue = Integer.parseInt(levelMap.get(majorityResult.getResult()));
                int weightLevelValue = Integer.parseInt(levelMap.get(weightResult.getResult()));
                // 值越小，档位越大
                if (majorityLevelValue < weightLevelValue) {
                    return majorityResult;
                } else {
                    return weightResult;
                }
            } else {
                return majorityResult;
            }
        }
        // 无过半档位或者未开启过半定档
        else {
            // 仲裁档位差
            int deviation = ParamCache.roughLevelConfigMap.get(String.valueOf(workId)).getDeviation();
            // 超过仲裁档位差
            Set<Integer> sources = markTasks.stream().map(MarkTaskRoughLevel::getLevelValue).collect(Collectors.toSet());
            if (ArbitrationService.overDeviation(deviation, sources)) {
                boolean autoCallBack = ParamCache.roughLevelConfigMap.get(String.valueOf(workId)).getAutoCallback() == 1;
                // 开启自动打回
                if (autoCallBack) {
                    int cumulateError = ParamCache.roughLevelConfigMap.get(String.valueOf(workId)).getCumulativeError();
                    List<String> markerIds = ArbitrationService.calcCumulateRoughLevel(cumulateError, markTasks);
                    // 不是所有评卷员累计误差都大于累计误差值
                    if (!markerIds.isEmpty() && markerIds.size() != markTasks.size()) {
                        return new RoughDeterResult(null, DeterType.REJECT, markerIds);
                    }
                }
                // 打回科组长
                return new RoughDeterResult(null, DeterType.ARBITRATE);
            } else {
                return calcWeight(workId, roughLevels, markTasks);
            }
        }
    }

    /**
     * 计算权重值
     *
     * @param markTasks 任务集合
     */
    private RoughDeterResult calcWeight(Long workId, List<RoughLevel> roughLevels, List<MarkTaskRoughLevel> markTasks) {
        // 是否开启去高去低再加权评卷（1：是， 0：否，默认0）
        boolean removeHighAndLow = Optional.ofNullable(ParamCache.roughLevelConfigMap.get(String.valueOf(String.valueOf(workId))).getRemoveHighAndLow()).orElse(0) == 1;
        if (removeHighAndLow && markTasks.size() > 2) {
            markTasks.sort((Comparator.comparing(MarkTaskRoughLevel::getResult)));
            markTasks.remove(0);//去掉最大值
            markTasks.remove(markTasks.size() - 1);//去掉最小值
        }

        MarkTaskRoughLevel[] tasks = markTasks.toArray(new MarkTaskRoughLevel[0]);
        //通过权重计算，平均值四舍五入
        double[] values = new double[tasks.length];
        double[] weights = new double[tasks.length];
        for (int i = 0; i < tasks.length; i++) {
            String levelResult = tasks[i].getResult();
            RoughLevel level = roughLevels.stream().filter(m -> m.getRoughCode().equals(levelResult)).findFirst().orElse(null);
            if (level == null) {
                throw new RuntimeException("无此档位");
            }
            MarkUser marker = markUserRepo.findOne(tasks[i].getMarkerId());
            double markerWeight = marker.getWeight();
            values[i] = level.getWeight() * markerWeight;
            weights[i] = markerWeight;
        }
        int avg = (int) Math.round(DoubleStream.of(values).sum() / DoubleStream.of(weights).sum());
        Optional<RoughLevel> levelOptional = roughLevels.stream().filter(o -> o.getMaxScore() >= avg && o.getMinScore() <= avg).findFirst();
        RoughLevel level = levelOptional.orElse(null);
        if (level == null) {
            roughLevels = roughLevels.stream().sorted(Comparator.comparing(RoughLevel::getMaxScore).reversed()).collect(Collectors.toList());
            if (avg > roughLevels.get(0).getMaxScore()) {
                level = roughLevels.get(0);
            }
            roughLevels = roughLevels.stream().sorted(Comparator.comparing(RoughLevel::getMinScore)).collect(Collectors.toList());
            if (avg < roughLevels.get(0).getMinScore()) {
                level = roughLevels.get(0);
            }
        }
        return new RoughDeterResult(level.getRoughCode(), DeterType.WEIGHT);
    }

    /**
     * 计算过半定档值
     *
     * @param tasks 任务集合
     */
    private RoughDeterResult calcMajority(MarkTaskRoughLevel[] tasks) {
        //如果相同的数量大于半数，直接定档
        Stream<MarkTaskRoughLevel> taskStream = Stream.of(tasks);
        Optional<Map.Entry<String, Long>> optional = taskStream
                .collect(Collectors.groupingBy(MarkTaskRoughLevel::getResult, Collectors.counting()))
                .entrySet()
                .stream()
                .max(Comparator.comparing(Map.Entry::getValue));
        if (optional.isPresent()) {
            Map.Entry<String, Long> entry = optional.get();
            if (entry.getValue() > tasks.length / 2) {
                return new RoughDeterResult(entry.getKey(), DeterType.MAJORITY);
            }
        }
        return null;
    }


    public static List<RoughLevel> parseRoughLevels(List<Level> levels){
        List<RoughLevel> roughLevels = new ArrayList<>();

        Map<String, List<Level>> listMap = levels.stream().collect(Collectors.groupingBy(Level::getRoughCode));
        for (Map.Entry<String, List<Level>> entry : listMap.entrySet()) {
            RoughLevel roughLevel = new RoughLevel();
            roughLevel.setRoughCode(entry.getKey());

            List<Level> levelList = entry.getValue();
            roughLevel.setWeight(levelList.get(0).getRoughWeight());
            int minScore = levelList.stream().min(Comparator.comparing(Level::getMinScore)).get().getMinScore();
            roughLevel.setMinScore(minScore);
            int maxScore = levelList.stream().max(Comparator.comparing(Level::getMaxScore)).get().getMaxScore();
            roughLevel.setMaxScore(maxScore);
            roughLevels.add(roughLevel);
        }
        return roughLevels;
    }
}

class RoughDeterResult {
    private String result;
    private DetermineRoughLevelService.DeterType deterType;
    private List<String> rejectTasks;

    public RoughDeterResult(String result, DetermineRoughLevelService.DeterType deterType) {
        this.result = result;
        this.deterType = deterType;
    }

    public RoughDeterResult(String result, DetermineRoughLevelService.DeterType deterType, List<String> rejectTasks) {
        this.result = result;
        this.deterType = deterType;
        this.rejectTasks = rejectTasks;
    }

    public String getResult() {
        return result;
    }

    public void setResult(String result) {
        this.result = result;
    }

    public DetermineRoughLevelService.DeterType getDeterType() {
        return deterType;
    }

    public void setDeterType(DetermineRoughLevelService.DeterType deterType) {
        this.deterType = deterType;
    }

    public List<String> getRejectTasks() {
        return rejectTasks;
    }

    public void setRejectTasks(List<String> rejectTasks) {
        this.rejectTasks = rejectTasks;
    }
}